local u = BittensGlobalTables.GetTable("BittensUtilities")
if u.SkipOrUpgrade(u, "Options Panels", 2) then
	return
end

local _G = _G
local CreateFrame = CreateFrame
local InterfaceOptions_AddCategory = InterfaceOptions_AddCategory
local InterfaceOptionsFrame = InterfaceOptionsFrame
local InterfaceOptionsFrame_OpenToCategory = InterfaceOptionsFrame_OpenToCategory
local InterfaceOptionsFrame_Show = InterfaceOptionsFrame_Show
local InterfaceOptionsFrameCancel_OnClick = InterfaceOptionsFrameCancel_OnClick
local UIDropDownMenu_AddButton = UIDropDownMenu_AddButton
local UIDropDownMenu_CreateInfo = UIDropDownMenu_CreateInfo
local UIDropDownMenu_Initialize = UIDropDownMenu_Initialize
local UIDropDownMenu_SetText = UIDropDownMenu_SetText
local UIDropDownMenu_SetWidth = UIDropDownMenu_SetWidth
local ipairs = ipairs
local math = math
local type = type
local wipe = wipe

local function createOption(name, default, frameType, template)
	return {
		Name = name,
		Default = default,
		FrameType = frameType,
		Template = template,
	}
end

function u.CreateCheckBoxOption(name, text, default)
	local option = createOption(
		name, default, "CheckButton", "InterfaceOptionsCheckButtonTemplate")
	option.Text = text
	return option
end

function u.CreateSliderOption(name, text, default, min, max, width)
	local option = createOption(
		name, default, "Slider", "OptionsSliderTemplate")
	option.Text = text
	option.Min = min
	option.Max = max
	option.Width = width
	return option
end

function u.CreateDropdownOption(name, default, width, ...)
	local option = createOption(
		name, default, "Frame", "UIDropDownMenuTemplate")
	option.Width = width
	option.Options = { ... }
	return option
end

-- Add the function apply() to the returned panel.  It, along with widgets in
-- the panel, should use/modify the values in panel.Settings.
--
-- You should make missing (nil) values in panel.Settings represent default
-- values.
--
-- If you provide a parent, they will share the same panel.Settings.  Call
-- Initialize() on the parent first, and you can omit the argument when calling
-- it on the children.
function u.CreateOptionsPanel(name, parent)	
	local panel = CreateFrame("Frame")
	panel.name = name
	if parent then
		panel.parent = parent.name
		panel.Settings = parent.Settings
		panel.Defaults = parent.Defaults
	else
		panel.Settings = { }
		panel.Defaults = { }
	end
	
	function panel.Initialize(savedSettingsTableName)
		if not parent then
			panel.SavedSettings = u.GetOrMakeTable(_G, savedSettingsTableName)
			u.ShallowCopy(panel.SavedSettings, panel.Settings)
		end
		for option, dropdown in u.Pairs(panel.UIDropDownMenuTemplate) do
			UIDropDownMenu_SetText(
				dropdown, panel.GetValue(option):format(panel.GetValue(option)))
		end
		panel.apply()
	end
	
	function panel.AddLabel(text, template, x, y)
		local label = panel:CreateFontString(nil, nil, template)
		label:SetPoint("TOPLEFT", x, y)
		label:SetText(text)
		return y - label:GetHeight()
	end
	
	function panel.Add(option, x, y)
		local widgetName = name .. option.Name -- TODO disambiguate via hierarchy?
		local widget = CreateFrame(
			option.FrameType, widgetName, panel, option.Template)
		
		local topPad = 0
		local bottomPad = 0
		if option.Template == "InterfaceOptionsCheckButtonTemplate" then
			_G[widgetName .. "Text"]:SetText(option.Text)
			widget:SetScript("OnClick", function()
				panel.Settings[option.Name] = not not widget:GetChecked()
				panel.apply()
			end)
		elseif option.Template == "OptionsSliderTemplate" then
			widget:SetWidth(option.Width)
			widget:SetMinMaxValues(option.Min, option.Max)
			_G[widgetName .. 'High']:SetText(option.Max);
			local tmpWidget = _G[widgetName .. 'Low']
			tmpWidget:SetText(option.Min);
			bottomPad = tmpWidget:GetHeight()
			tmpWidget = _G[widgetName .. "Text"]
			tmpWidget:SetText(" ")
			topPad = tmpWidget:GetHeight()
			widget:SetScript("OnValueChanged", function(self, value)
				value = math.floor(value)
				panel.Settings[option.Name] = value
				tmpWidget:SetText(option.Text:format(value))
				panel.apply()
			end)
		elseif option.Template == "UIDropDownMenuTemplate" then
			UIDropDownMenu_SetWidth(widget, option.Width)
			UIDropDownMenu_Initialize(widget, function()
				for _, text in ipairs(option.Options) do
					local info = UIDropDownMenu_CreateInfo()
					info.text = text
					info.func = function()
						panel.Settings[option.Name] = text
						panel.apply()
					end
					UIDropDownMenu_AddButton(info)
				end
			end)
		end
		
		widget:SetPoint("TOPLEFT", x, y - topPad)
		u.GetOrMakeTable(panel, option.Template)[option] = widget
		panel.Defaults[option.Name] = option.Default
		return y - widget:GetHeight() - topPad - bottomPad
	end
	
	function panel.AddCheckBoxGroup(label, options, x, y)
		y = panel.AddLabel(label, "GameFontNormalLarge", x, y)
		for _, option in ipairs(options) do
			y = panel.Add(option, x, y)
		end
		return y
	end
	
	function panel.GetValue(option)
		if type(option) == "table" then
			option = option.Name
		end
		local value = panel.Settings[option]
		if value == nil then
			return panel.Defaults[option]
		else
			return value
		end
	end
	
	function panel.default()
		wipe(panel.Settings)
		panel.apply()
	end
	
	function panel.refresh()
		for option, checkBox in u.Pairs(panel.InterfaceOptionsCheckButtonTemplate) do
			checkBox:SetChecked(panel.GetValue(option))
		end
		for option, slider in u.Pairs(panel.OptionsSliderTemplate) do
			slider:SetValue(panel.GetValue(option))
		end
		for option, dropdown in u.Pairs(panel.UIDropDownMenuTemplate) do
			UIDropDownMenu_SetText(dropdown, panel.GetValue(option))
		end
	end
	
	function panel.okay()
		if not parent then
			u.ShallowCopy(panel.Settings, panel.SavedSettings)
		end
		panel.apply()
	end
	
	function panel.cancel()
		if not parent then
			u.ShallowCopy(panel.SavedSettings, panel.Settings)
		end
		panel.apply()
	end
	
	InterfaceOptions_AddCategory(panel)
	return panel
end

function u.ToggleOptionsPanel(panel)
	if not InterfaceOptionsFrame:IsVisible() then
		InterfaceOptionsFrame_Show()
		InterfaceOptionsFrame_OpenToCategory(panel)
	elseif panel:IsVisible() then
		InterfaceOptionsFrameCancel_OnClick()
	else
		InterfaceOptionsFrame_OpenToCategory(panel)
	end
end
